# define directory as package

# flatten access
from ieegprep.bids.data_structure import list_bids_datasets
from ieegprep.bids.sidecars import load_channel_info, load_event_info, load_ieeg_sidecar
#from data_epoch import load_data_epochs, load_data_epochs_averages

